=== modified file 'NEWS'
--- NEWS	2008-10-30 03:56:00 +0000
+++ NEWS	2008-10-31 01:40:06 +0000
@@ -44,6 +44,17 @@
     * lp: lookups no longer include usernames, making them shareable and
       shorter. (Aaron Bentley)
 
+    * New ``PackRepository.autopack`` smart server RPC, which does
+      autopacking entirely on the server.  This is much faster than
+      autopacking via plain file methods, which downloads a large amount
+      of pack data and then re-uploads the same pack data into a single
+      file.  This fixes a major (although infrequent) cause of lengthy
+      delays when using a smart server.  For example, pushing the 10th
+      revision to a repository with 9 packs now takes 44 RPCs rather than
+      179, and much less bandwidth too.  This requires Bazaar 1.9 on both
+      the client and the server, otherwise the client will fallback to the
+      slower method.  (Andrew Bennetts)
+
   BUG FIXES:
 
     * A failure to load a plugin due to an IncompatibleAPI exception is

=== modified file 'bzrlib/remote.py'
--- bzrlib/remote.py	2008-10-30 15:20:00 +0000
+++ bzrlib/remote.py	2008-10-31 01:57:16 +0000
@@ -810,6 +810,8 @@
             other, self).search_missing_revision_ids(revision_id, find_ghosts)
 
     def fetch(self, source, revision_id=None, pb=None, find_ghosts=False):
+        # Not delegated to _real_repository so that InterRepository.get has a
+        # chance to find an InterRepository specialised for RemoteRepository.
         if self.has_same_location(source):
             # check that last_revision is in 'from' and then return a
             # no-operation.
@@ -820,7 +822,6 @@
         inter = repository.InterRepository.get(source, self)
         try:
             return inter.fetch(revision_id=revision_id, pb=pb, find_ghosts=find_ghosts)
-
         except NotImplementedError:
             raise errors.IncompatibleRepositories(source, self)
 
@@ -1217,6 +1218,23 @@
         count = str(recipe[2])
         return '\n'.join((start_keys, stop_keys, count))
 
+    def autopack(self):
+        path = self.bzrdir._path_for_remote_call(self._client)
+        try:
+            response = self._call('PackRepository.autopack', path)
+        except errors.UnknownSmartMethod:
+            self._ensure_real()
+            self._real_repository._pack_collection.autopack()
+            return
+        if self._real_repository is not None:
+            # Reset the real repository's cache of pack names.
+            # XXX: At some point we may be able to skip this and just rely on
+            # the automatic retry logic to do the right thing, but for now we
+            # err on the side of being correct rather than being optimal.
+            self._real_repository._pack_collection.reload_pack_names()
+        if response[0] != 'ok':
+            raise errors.UnexpectedSmartServerResponse(response)
+
 
 class RemoteBranchLockableFiles(LockableFiles):
     """A 'LockableFiles' implementation that talks to a smart server.

=== modified file 'bzrlib/repository.py'
--- bzrlib/repository.py	2008-10-30 15:20:00 +0000
+++ bzrlib/repository.py	2008-10-31 01:14:22 +0000
@@ -2800,7 +2800,6 @@
             fetcher = RepoFetcher(self.target, self.source, revision_id,
                                   pb, find_ghosts)
             return fetcher.count_copied, fetcher.failed_revisions
-        from bzrlib.repofmt.pack_repo import Packer
         mutter("Using fetch logic to copy between %s(%s) and %s(%s)",
                self.source, self.source._format, self.target, self.target._format)
         self.count_copied = 0
@@ -2814,7 +2813,8 @@
             revision_ids = source_revision_ids - \
                 frozenset(self.target_get_parent_map(source_revision_ids))
             revision_keys = [(revid,) for revid in revision_ids]
-            index = self.target._pack_collection.revision_index.combined_index
+            target_pack_collection = self._get_target_pack_collection()
+            index = target_pack_collection.revision_index.combined_index
             present_revision_ids = set(item[1][0] for item in
                 index.iter_entries(revision_keys))
             revision_ids = set(revision_ids) - present_revision_ids
@@ -2836,19 +2836,31 @@
                 raise errors.InstallFailed([revision_id])
             if len(revision_ids) == 0:
                 return (0, [])
-        packs = self.source._pack_collection.all_packs()
-        pack = Packer(self.target._pack_collection, packs, '.fetch',
+        return self._pack(self.source, self.target, revision_ids)
+
+    def _pack(self, source, target, revision_ids):
+        from bzrlib.repofmt.pack_repo import Packer
+        target_pack_collection = self._get_target_pack_collection()
+        packs = source._pack_collection.all_packs()
+        pack = Packer(target_pack_collection, packs, '.fetch',
             revision_ids).pack()
         if pack is not None:
-            self.target._pack_collection._save_pack_names()
+            target_pack_collection._save_pack_names()
+            copied_revs = pack.get_revision_count()
             # Trigger an autopack. This may duplicate effort as we've just done
             # a pack creation, but for now it is simpler to think about as
             # 'upload data, then repack if needed'.
-            self.target._pack_collection.autopack()
-            return (pack.get_revision_count(), [])
+            self._autopack()
+            return (copied_revs, [])
         else:
             return (0, [])
 
+    def _autopack(self):
+        self.target._pack_collection.autopack()
+        
+    def _get_target_pack_collection(self):
+        return self.target._pack_collection
+
     @needs_read_lock
     def search_missing_revision_ids(self, revision_id=None, find_ghosts=True):
         """See InterRepository.missing_revision_ids().
@@ -3052,6 +3064,9 @@
 
 
 class InterOtherToRemote(InterRepository):
+    """An InterRepository that simply delegates to the 'real' InterRepository
+    calculated for (source, target._real_repository).
+    """
 
     _walk_to_common_revisions_batch_size = 50
 
@@ -3127,6 +3142,39 @@
 
 
 
+class InterPackToRemotePack(InterPackRepo):
+    """A specialisation of InterPackRepo for a target that is a
+    RemoteRepository.
+
+    This will use the get_parent_map RPC rather than plain readvs, and also
+    uses an RPC for autopacking.
+    """
+
+    _walk_to_common_revisions_batch_size = 50
+
+    @staticmethod
+    def is_compatible(source, target):
+        from bzrlib.repofmt.pack_repo import RepositoryFormatPack
+        if isinstance(source._format, RepositoryFormatPack):
+            if isinstance(target, remote.RemoteRepository):
+                target._ensure_real()
+                if isinstance(target._real_repository._format,
+                              RepositoryFormatPack):
+                    if InterRepository._same_model(source, target):
+                        return True
+        return False
+    
+    def _autopack(self):
+        self.target.autopack()
+        
+    def _get_target_pack_collection(self):
+        return self.target._real_repository._pack_collection
+
+    @classmethod
+    def _get_repo_format_to_test(self):
+        return None
+
+
 InterRepository.register_optimiser(InterDifferingSerializer)
 InterRepository.register_optimiser(InterSameDataRepository)
 InterRepository.register_optimiser(InterWeaveRepo)
@@ -3136,6 +3184,7 @@
 InterRepository.register_optimiser(InterPackRepo)
 InterRepository.register_optimiser(InterOtherToRemote)
 InterRepository.register_optimiser(InterRemoteToOther)
+InterRepository.register_optimiser(InterPackToRemotePack)
 
 
 class CopyConverter(object):

=== added file 'bzrlib/smart/packrepository.py'
--- bzrlib/smart/packrepository.py	1970-01-01 00:00:00 +0000
+++ bzrlib/smart/packrepository.py	2008-10-31 01:57:16 +0000
@@ -0,0 +1,45 @@
+# Copyright (C) 2008 Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+"""Server-side pack repository related request implmentations."""
+
+from bzrlib.smart.request import (
+    FailedSmartServerResponse,
+    SuccessfulSmartServerResponse,
+    )
+
+
+from bzrlib.smart.repository import (
+    SmartServerRepositoryRequest,
+    )
+
+
+class SmartServerPackRepositoryAutopack(SmartServerRepositoryRequest):
+
+    def do_repository_request(self, repository):
+        pack_collection = getattr(repository, '_pack_collection', None)
+        if pack_collection is None:
+            # This is a not a pack repo, so asking for an autopack is just a
+            # no-op.
+            return SuccessfulSmartServerResponse(('ok',))
+        repository.lock_write()
+        try:
+            repository._pack_collection.autopack()
+        finally:
+            repository.unlock()
+        return SuccessfulSmartServerResponse(('ok',))
+
+

=== modified file 'bzrlib/smart/request.py'
--- bzrlib/smart/request.py	2008-10-16 05:54:43 +0000
+++ bzrlib/smart/request.py	2008-10-30 03:47:59 +0000
@@ -443,6 +443,9 @@
     'readv', 'bzrlib.smart.vfs', 'ReadvRequest')
 request_handlers.register_lazy(
     'rename', 'bzrlib.smart.vfs', 'RenameRequest')
+request_handlers.register_lazy(
+    'PackRepository.autopack', 'bzrlib.smart.packrepository',
+    'SmartServerPackRepositoryAutopack')
 request_handlers.register_lazy('Repository.gather_stats',
                                'bzrlib.smart.repository',
                                'SmartServerRepositoryGatherStats')

=== modified file 'bzrlib/tests/test_pack_repository.py'
--- bzrlib/tests/test_pack_repository.py	2008-10-23 19:39:59 +0000
+++ bzrlib/tests/test_pack_repository.py	2008-10-31 01:37:43 +0000
@@ -37,6 +37,10 @@
     upgrade,
     workingtree,
     )
+from bzrlib.smart import (
+    client,
+    server,
+    )
 from bzrlib.tests import (
     TestCase,
     TestCaseWithTransport,
@@ -627,6 +631,44 @@
         self.assertTrue(large_pack_name in pack_names)
 
 
+class TestSmartServerAutopack(TestCaseWithTransport):
+
+    def setUp(self):
+        super(TestSmartServerAutopack, self).setUp()
+        # Create a smart server that publishes whatever the backing VFS server
+        # does.
+        self.smart_server = server.SmartTCPServer_for_testing()
+        self.smart_server.setUp(self.get_server())
+        self.addCleanup(self.smart_server.tearDown)
+        # Log all HPSS calls into self.hpss_calls.
+        client._SmartClient.hooks.install_named_hook(
+            'call', self.capture_hpss_call, None)
+        self.hpss_calls = []
+
+    def capture_hpss_call(self, params):
+        self.hpss_calls.append(params.method)
+
+    def get_format(self):
+        return bzrdir.format_registry.make_bzrdir(self.format_name)
+
+    def test_autopack_rpc_is_used_when_using_hpss(self):
+        # Make local and remote repos
+        tree = self.make_branch_and_tree('local', format=self.get_format())
+        self.make_branch_and_tree('remote', format=self.get_format())
+        remote_branch_url = self.smart_server.get_url() + 'remote'
+        remote_branch = bzrdir.BzrDir.open(remote_branch_url).open_branch()
+        # Make 9 local revisions, and push them one at a time to the remote
+        # repo to produce 9 pack files.
+        for x in range(9):
+            tree.commit('commit %s' % x)
+            tree.branch.push(remote_branch)
+        # Make one more push to trigger an autopack
+        self.hpss_calls = []
+        tree.commit('commit triggering pack')
+        tree.branch.push(remote_branch)
+        self.assertTrue('PackRepository.autopack' in self.hpss_calls)
+
+
 def load_tests(basic_tests, module, test_loader):
     # these give the bzrdir canned format name, and the repository on-disk
     # format string

=== modified file 'bzrlib/tests/test_remote.py'
--- bzrlib/tests/test_remote.py	2008-10-29 01:16:22 +0000
+++ bzrlib/tests/test_remote.py	2008-10-31 01:57:16 +0000
@@ -171,14 +171,7 @@
     """Lookalike for _SmartClient allowing testing."""
     
     def __init__(self, fake_medium_base='fake base'):
-        """Create a FakeClient.
-
-        :param responses: A list of response-tuple, body-data pairs to be sent
-            back to callers.  A special case is if the response-tuple is
-            'unknown verb', then a UnknownSmartMethod will be raised for that
-            call, using the second element of the tuple as the verb in the
-            exception.
-        """
+        """Create a FakeClient."""
         self.responses = []
         self._calls = []
         self.expecting_body = False
@@ -1450,6 +1443,73 @@
         src_repo.copy_content_into(dest_repo)
 
 
+class _StubRealPackRepository(object):
+
+    def __init__(self, calls):
+        self._pack_collection = _StubPackCollection(calls)
+
+
+class _StubPackCollection(object):
+
+    def __init__(self, calls):
+        self.calls = calls
+
+    def autopack(self):
+        self.calls.append(('pack collection autopack',))
+
+    def reload_pack_names(self):
+        self.calls.append(('pack collection reload_pack_names',))
+
+    
+class TestRemotePackRepositoryAutoPack(TestRemoteRepository):
+    """Tests for RemoteRepository.autopack implementation."""
+
+    def test_ok(self):
+        """When the server returns 'ok' and there's no _real_repository, then
+        nothing else happens: the autopack method is done.
+        """
+        transport_path = 'quack'
+        repo, client = self.setup_fake_client_and_repository(transport_path)
+        client.add_expected_call(
+            'PackRepository.autopack', ('quack/',), 'success', ('ok',))
+        repo.autopack()
+        client.finished_test()
+
+    def test_ok_with_real_repo(self):
+        """When the server returns 'ok' and there is a _real_repository, then
+        the _real_repository's reload_pack_name's method will be called.
+        """
+        transport_path = 'quack'
+        repo, client = self.setup_fake_client_and_repository(transport_path)
+        client.add_expected_call(
+            'PackRepository.autopack', ('quack/',),
+            'success', ('ok',))
+        repo._real_repository = _StubRealPackRepository(client._calls)
+        repo.autopack()
+        self.assertEqual(
+            [('call', 'PackRepository.autopack', ('quack/',)),
+             ('pack collection reload_pack_names',)],
+            client._calls)
+        
+    def test_backwards_compatibility(self):
+        """If the server does not recognise the PackRepository.autopack verb,
+        fallback to the real_repository's implementation.
+        """
+        transport_path = 'quack'
+        repo, client = self.setup_fake_client_and_repository(transport_path)
+        client.add_unknown_method_response('PackRepository.autopack')
+        def stub_ensure_real():
+            client._calls.append(('_ensure_real',))
+            repo._real_repository = _StubRealPackRepository(client._calls)
+        repo._ensure_real = stub_ensure_real
+        repo.autopack()
+        self.assertEqual(
+            [('call', 'PackRepository.autopack', ('quack/',)),
+             ('_ensure_real',),
+             ('pack collection autopack',)],
+            client._calls)
+
+
 class TestErrorTranslationBase(tests.TestCaseWithMemoryTransport):
     """Base class for unit tests for bzrlib.remote._translate_error."""
 
@@ -1483,7 +1543,7 @@
                 **context)
         return translated_error
 
-    
+
 class TestErrorTranslationSuccess(TestErrorTranslationBase):
     """Unit tests for bzrlib.remote._translate_error.
     
@@ -1595,7 +1655,6 @@
         self.assertEqual(expected_error, translated_error)
 
 
-
 class TestErrorTranslationRobustness(TestErrorTranslationBase):
     """Unit tests for bzrlib.remote._translate_error's robustness.
     

=== modified file 'bzrlib/tests/test_repository.py'
--- bzrlib/tests/test_repository.py	2008-10-28 17:41:35 +0000
+++ bzrlib/tests/test_repository.py	2008-10-29 21:39:27 +0000
@@ -473,6 +473,7 @@
 class DummyRepository(object):
     """A dummy repository for testing."""
 
+    _format = None
     _serializer = None
 
     def supports_rich_root(self):

=== modified file 'bzrlib/tests/test_smart.py'
--- bzrlib/tests/test_smart.py	2008-09-15 09:42:16 +0000
+++ bzrlib/tests/test_smart.py	2008-10-31 01:57:16 +0000
@@ -1037,6 +1037,53 @@
             SmartServerResponse(('yes',)), response)
 
 
+class TestSmartServerPackRepositoryAutopack(tests.TestCaseWithTransport):
+
+    def make_repo_needing_autopacking(self, path='.'):
+        # Make a repo in need of autopacking.
+        tree = self.make_branch_and_tree('.', format='pack-0.92')
+        repo = tree.branch.repository
+        # monkey-patch the pack collection to disable autopacking
+        repo._pack_collection._max_pack_count = lambda count: count
+        for x in range(10):
+            tree.commit('commit %s' % x)
+        self.assertEqual(10, len(repo._pack_collection.names()))
+        del repo._pack_collection._max_pack_count
+        return repo
+
+    def test_autopack_needed(self):
+        repo = self.make_repo_needing_autopacking()
+        backing = self.get_transport()
+        request = smart.packrepository.SmartServerPackRepositoryAutopack(
+            backing)
+        response = request.execute('')
+        self.assertEqual(SmartServerResponse(('ok',)), response)
+        repo._pack_collection.reload_pack_names()
+        self.assertEqual(1, len(repo._pack_collection.names()))
+    
+    def test_autopack_not_needed(self):
+        tree = self.make_branch_and_tree('.', format='pack-0.92')
+        repo = tree.branch.repository
+        for x in range(9):
+            tree.commit('commit %s' % x)
+        backing = self.get_transport()
+        request = smart.packrepository.SmartServerPackRepositoryAutopack(
+            backing)
+        response = request.execute('')
+        self.assertEqual(SmartServerResponse(('ok',)), response)
+        repo._pack_collection.reload_pack_names()
+        self.assertEqual(9, len(repo._pack_collection.names()))
+    
+    def test_autopack_on_nonpack_format(self):
+        """A request to autopack a non-pack repo is a no-op."""
+        repo = self.make_repository('.', format='knit')
+        backing = self.get_transport()
+        request = smart.packrepository.SmartServerPackRepositoryAutopack(
+            backing)
+        response = request.execute('')
+        self.assertEqual(SmartServerResponse(('ok',)), response)
+        
+
 class TestHandlers(tests.TestCase):
     """Tests for the request.request_handlers object."""
 
@@ -1083,6 +1130,9 @@
             smart.request.request_handlers.get('BzrDir.open_branch'),
             smart.bzrdir.SmartServerRequestOpenBranch)
         self.assertEqual(
+            smart.request.request_handlers.get('PackRepository.autopack'),
+            smart.packrepository.SmartServerPackRepositoryAutopack)
+        self.assertEqual(
             smart.request.request_handlers.get('Repository.gather_stats'),
             smart.repository.SmartServerRepositoryGatherStats)
         self.assertEqual(

